# Copyright (c) 2023 K.Y. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from __future__ import annotations

import asyncio
import io
import os
import traceback
from typing import AsyncGenerator

import msgspec
from litestar import MediaType, Request, get, post, status_codes
from litestar.datastructures import UploadFile
from litestar.enums import RequestEncodingType
from litestar.exceptions import HTTPException
from litestar.params import Body
from litestar.response import Stream
from rich import print
from typing_extensions import Annotated

from ..pack import encode
from .interface import (
    AttachRequest,
    DetachRequest,
    StructDeleteBatchData,
    StructGetBatchData,
    StructSetBatchData,
    StructSetData,
)
from .manager import DBManager

# from litestar.serialization import decode_json, decode_msgpack


_db_manager = DBManager(
    root_path="./FLAXKV_DB",
    raw_mode=True,
    log_level=os.environ.get("FLAXKV_LOG_LEVEL", "INFO"),
)


def _get_db(db_name: str):
    db = _db_manager.get(db_name)
    if db is None:
        raise HTTPException(status_code=500, detail="db not found")
    return db


@get(
    "/healthz",
    summary="Perform a Health Check",
    response_description="Return HTTP Status Code 200 (OK)",
    status_code=status_codes.HTTP_200_OK,
    media_type=MediaType.TEXT,
)
async def healthz() -> str:
    return "OK"


@get("/check_db")
async def check_db(db_name: str) -> bool:
    return _get_db(db_name) is not None


@get("/disconnect")
async def disconnect(client_id: str) -> dict:
    try:
        _db_manager.subscribers[client_id]['disconnect_event'].set()
        return {"success": True}
    except KeyError:
        ...
    except Exception as e:
        traceback.print_exc()
        raise HTTPException(status_code=500, detail=str(e))


@post(path="/connect")
async def connect(data: AttachRequest) -> Stream:
    try:
        db = _db_manager.get(data.db_name)
        q = asyncio.Queue()
        client = {
            "db_name": data.db_name,
            "update_data": q,
            "disconnect_event": asyncio.Event(),
        }
        client_id = data.client_id
        _db_manager.subscribers[client_id] = client

        print(f"{' Client connected ':*^60}")
        print(f"{client_id=}")
        print(f"Current subscribers: {_db_manager.subscribers.keys()}")

        if db is None:
            _db_manager.set_db(
                db_name=data.db_name, backend=data.backend, rebuild=data.rebuild
            )
        elif data.rebuild:
            db.clear(wait=True)
            _db_manager.set_db(
                db_name=data.db_name, backend=data.backend, rebuild=data.rebuild
            )

        async def stream(client: dict) -> AsyncGenerator:
            try:
                chunk_size = 1024 * 1024
                while True:
                    done, pending = await asyncio.wait(
                        [
                            asyncio.create_task(q.get()),
                            asyncio.create_task(client['disconnect_event'].wait()),
                        ],
                        return_when=asyncio.FIRST_COMPLETED,
                    )
                    if client['disconnect_event'].is_set():
                        for task in pending:
                            task.cancel()
                        break

                    else:
                        data_dict = done.pop().result()
                        # print(f"{data_dict=}")
                        # data_dict = await q.get()
                        assert client_id != data_dict['client_id']
                        buffer_dict = data_dict.get("buffer_dict")
                        if buffer_dict:
                            bytes_data = msgspec.msgpack.encode(
                                {
                                    "type": 'buffer_dict',
                                    'data': buffer_dict,
                                    'time': data_dict['time'],
                                }
                            )
                        else:
                            delete_list = data_dict['delete_keys']
                            bytes_data = msgspec.msgpack.encode(
                                {
                                    "type": 'delete_keys',
                                    'data': {key: b"" for key in delete_list},
                                    'time': data_dict['time'],
                                }
                            )

                        for i in range(0, len(bytes_data), chunk_size):
                            yield bytes_data[i : i + chunk_size]
                        yield b"data: end\n\n"
            finally:
                if isinstance(pending, set):
                    for task in pending:
                        task.cancel()
                print(f"{' Client disconnected ':*^60}")
                _db_manager.subscribers.pop(data.client_id)
                print(f"Current subscribers: {_db_manager.subscribers.keys()}")

        return Stream(stream(client))

    except Exception as e:

        async def stream_error(e):

            yield bytes(f"error {data.client_id} {e}\n", "utf-8")

        traceback.print_exc()
        return Stream(stream_error(e))


@post(path="/detach")
async def detach(data: DetachRequest) -> dict:
    # todo switch `post` to `get`
    db = _db_manager.detach(db_name=data.db_name)
    if db is None:
        return {"success": False, "info": "db not found"}
    return {"success": True}


@post(path="/set")
async def set_(db_name: str, request: Request) -> None:
    db = _get_db(db_name)
    data = await request.body()
    try:
        data = msgspec.msgpack.decode(data, type=StructSetData)
    except Exception as e:
        traceback.print_exc()
        raise HTTPException(status_code=500, detail=str(e))
    db[data.key] = data.value


@post("/set_batch")
async def set_batch_(db_name: str, request: Request) -> None:
    db = _get_db(db_name)
    data = await request.body()
    try:
        data = msgspec.msgpack.decode(data, type=StructSetBatchData)
        for key, value in data.data.items():
            db[key] = value
    except Exception as e:
        traceback.print_exc()
        raise HTTPException(status_code=500, detail=str(e))


@post("/set_batch_stream", media_type=MediaType.TEXT)
async def set_batch_stream_(
    data: Annotated[UploadFile, Body(media_type=RequestEncodingType.MULTI_PART)],
) -> None:
    # litestar >= 2.5.0 fixed: https://github.com/litestar-org/litestar/issues/2939
    raw_data = await data.read()
    db_name = data.filename
    db = _get_db(db_name)

    try:
        content = msgspec.msgpack.decode(raw_data, type=StructSetBatchData)
        assert db_name == _db_manager.subscribers[content.client_id]['db_name']
        clients = [
            client
            for client_id, client in _db_manager.subscribers.items()
            if client_id != content.client_id
        ]
        # print(f"{clients=}")
        for key, value in content.data.items():
            db[key] = value
        for client in clients:
            await client['update_data'].put(
                {
                    "buffer_dict": content.data,
                    'time': content.time,
                    "client_id": content.client_id,
                }
            )

    except Exception as e:
        traceback.print_exc()
        raise HTTPException(status_code=500, detail=str(e))


@post("/get", media_type=MediaType.TEXT)
async def get_(db_name: str, request: Request) -> bytes:
    db = _get_db(db_name)
    key = await request.body()
    value = db.get(key, None)
    if value is None:
        return b'iamnull123'
    return value


@post("/get_batch", media_type=MediaType.TEXT)
async def get_batch_(db_name: str, request: Request) -> bytes:
    db = _get_db(db_name)
    data = await request.body()
    data = msgspec.msgpack.decode(data, type=StructGetBatchData)
    values = db.get_batch(data.keys)
    return encode(values)


@post("/delete")
async def delete_(db_name: str, request: Request) -> None:
    db = _get_db(db_name)
    key = await request.body()
    try:
        db.pop(key)
    except Exception as e:
        traceback.print_exc()
        raise HTTPException(status_code=500, detail=str(e))


@post("/delete_batch")
async def delete_batch_(db_name: str, request: Request) -> None:
    db = _get_db(db_name)
    data = await request.body()
    try:

        data = msgspec.msgpack.decode(data, type=StructDeleteBatchData)
        assert db_name == _db_manager.subscribers[data.client_id]['db_name']

        for key in data.keys:
            db.pop(key)

        clients = [
            client
            for client_id, client in _db_manager.subscribers.items()
            if client_id != data.client_id
        ]
        for client in clients:
            await client['update_data'].put(
                {
                    "delete_keys": data.keys,
                    'time': data.time,
                    "client_id": data.client_id,
                }
            )

    except Exception as e:
        traceback.print_exc()
        raise HTTPException(status_code=500, detail=str(e))


@get("/keys", media_type=MediaType.TEXT)
async def keys_(db_name: str) -> bytes:
    db = _get_db(db_name)
    try:
        return encode(list(db.keys()))
    except Exception as e:
        traceback.print_exc()
        raise HTTPException(status_code=500, detail=str(e))


@get("/dict", media_type=MediaType.TEXT)
async def dict_(db_name: str) -> bytes:
    db = _get_db(db_name)
    try:
        return encode(db.db_dict())
    except Exception as e:
        traceback.print_exc()
        raise HTTPException(status_code=500, detail=str(e))


async def stream_generator(
    data: bytes, chunk_size=1024 * 1024
) -> AsyncGenerator[bytes, None]:
    with io.BytesIO(data) as data_io:
        while chunk := data_io.read(chunk_size):
            yield chunk


@get("/keys_stream", media_type=MediaType.TEXT)
async def keys_stream_(db_name: str) -> Stream:
    db = _get_db(db_name)
    try:
        result_bin = encode(list(db.keys()))
        return Stream(stream_generator(result_bin))
    except Exception as e:
        traceback.print_exc()
        raise HTTPException(status_code=500, detail=str(e))


@post("/get_batch_stream", media_type=MediaType.TEXT)
async def get_batch_stream_(db_name: str, request: Request) -> Stream:
    db = _get_db(db_name)
    data = await request.body()
    try:
        data = msgspec.msgpack.decode(data, type=StructGetBatchData)
        values = db.get_batch(data.keys)
        result_bin = encode(values)
        return Stream(stream_generator(result_bin))
    except Exception as e:
        traceback.print_exc()
        raise HTTPException(status_code=500, detail=str(e))


@get("/dict_stream", media_type=MediaType.TEXT)
async def dict_stream_(db_name: str) -> Stream:
    db = _get_db(db_name)
    try:
        result_bin = encode(db.db_dict())
        return Stream(stream_generator(result_bin))
    except Exception as e:
        traceback.print_exc()
        raise HTTPException(status_code=500, detail=str(e))


@get("/stat", media_type=MediaType.TEXT)
async def stat_(db_name: str) -> bytes:
    db = _get_db(db_name)
    try:
        return encode(db.stat())
    except Exception as e:
        traceback.print_exc()
        raise HTTPException(status_code=500, detail=str(e))



# Copyright (c) 2023 K.Y. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from __future__ import annotations

from litestar import Litestar
from litestar.openapi import OpenAPIConfig

from .. import __version__
from .route import *


def on_startup():
    ...


def on_shutdown():
    ...


app = Litestar(
    route_handlers=[
        healthz,
        connect,
        disconnect,
        detach,
        get_,
        get_batch_stream_,
        set_,
        set_batch_,
        set_batch_stream_,
        delete_,
        delete_batch_,
        keys_,
        keys_stream_,
        dict_,
        dict_stream_,
        stat_,
        check_db,
    ],
    on_startup=[on_startup],
    on_shutdown=[on_shutdown],
    openapi_config=OpenAPIConfig(title="FlaxKV", version=f"v{__version__}"),
)



# Copyright (c) 2023 K.Y. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import os
import platform

import fire

try:
    import uvloop

    uvloop.install()
except:
    ...


class Cli:
    @staticmethod
    def run(port=8000, **kwargs):
        """
        Runs the application using the Uvicorn server.

        Args:
            port (int): The port number on which to run the server. Default is 8000.

        Returns:
            None
        """

        if platform.system() == "Windows":
            os.environ["TZ"] = ""

        log_level = kwargs.get("log", "info")
        os.environ['FLAXKV_LOG_LEVEL'] = log_level.upper()

        http2 = kwargs.get("http2", False)
        if http2:
            print("use http2")
            from hypercorn.asyncio import serve
            from hypercorn.config import Config

            from flaxkv.serve.app import app

            config = Config()
            config.bind = [f"0.0.0.0:{port}"]
            asyncio.run(serve(app, config))
        else:
            import uvicorn

            uvicorn.run(
                app="flaxkv.serve.app:app",
                host=kwargs.get("host", "0.0.0.0"),
                port=port,
                workers=1,
                app_dir="..",
                ssl_keyfile=kwargs.get("ssl_keyfile", None),
                ssl_certfile=kwargs.get("ssl_certfile", None),
                use_colors=True,
                log_level=log_level.lower(),
            )


def main():
    fire.Fire(Cli)


if __name__ == "__main__":
    main()



# Copyright (c) 2023 K.Y. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from __future__ import annotations

import logging
import time
from functools import wraps
from typing import TYPE_CHECKING

from rich import print
from rich.text import Text

if TYPE_CHECKING:
    from flaxkv import FlaxKV

ENABLED_MEASURE_TIME_DECORATOR = True


def class_measure_time(logger=None, level=logging.INFO, prec=3):
    def decorate(func):
        """Log the runtime of the decorated function."""

        @wraps(func)
        def wrapper(self, *args, **kwargs):
            if not ENABLED_MEASURE_TIME_DECORATOR:
                return func(self, *args, **kwargs)
            start = time.perf_counter()
            value = func(self, *args, **kwargs)
            end = time.perf_counter()
            cost_time = end - start
            time_str = f"{cost_time:.{int(prec)}E}"
            msg = f"{func.__name__}:{self._db_manager.db_type}"
            if logger:
                show_string = f"Finished {msg} in {time_str} secs."
                logger.log(level, show_string)
            else:
                rgb_cost_time = Text(time_str, style='green')
                rgb_msg = Text(f"{msg}", style="cyan")
                str_tuple = (f"Finished", rgb_msg, "in", rgb_cost_time, "secs.")
                print(*str_tuple, sep=' ')
            return value

        return wrapper

    return decorate


def msg_encoder(func):
    from .pack import encode

    @wraps(func)
    async def wrapper(*args, **kwargs):
        result = await func(*args, **kwargs)
        return encode(result)

    return wrapper


def retry(max_retries=3, delay=1, backoff=2, exceptions=(Exception,)):
    """
    A decorator for automatically retrying a function upon encountering specified exceptions.

    Args:
        max_retries (int): The maximum number of times to retry the function.
        delay (float): The initial delay between retries in seconds.
        backoff (float): The multiplier by which the delay should increase after each retry.
        exceptions (tuple): A tuple of exception classes upon which to retry.

    Returns:
        The return value of the wrapped function, if it succeeds.
        Raises the last encountered exception if the function never succeeds.
    """

    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            retries = 0
            current_delay = delay
            while retries <= max_retries:
                try:
                    return func(*args, **kwargs)
                except exceptions as e:
                    retries += 1
                    if retries == max_retries:
                        raise
                    print(
                        f"Retrying `{func.__name__}` after {current_delay} seconds, retry : {retries}\n"
                    )
                    time.sleep(current_delay)
                    current_delay *= backoff

        return wrapper

    return decorator


def cache(db: FlaxKV = None):
    """Keep a cache of previous function calls."""

    if db is None:
        db = {}

    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            key = (args, tuple(sorted(kwargs.items())))
            if key in db:
                return db[key]
            result = func(*args, **kwargs)
            db[key] = result
            return result

        return wrapper

    return decorator


def singleton(cls):
    instances = {}

    @wraps(cls)
    def get_instance(*args, **kwargs):
        if cls not in instances:
            instances[cls] = cls(*args, **kwargs)
        return instances[cls]

    return get_instance



# Copyright (c) 2023 K.Y. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from __future__ import annotations

from dataclasses import dataclass
from typing import Dict, List

import msgspec
from litestar.enums import RequestEncodingType
from litestar.params import Body


@dataclass
class AttachRequest:
    db_name: str
    backend: str
    rebuild: bool
    client_id: str


@dataclass
class DetachRequest:
    db_name: str


class StructSetData(msgspec.Struct):
    key: bytes
    value: bytes


class StructSetBatchData(msgspec.Struct):
    data: Dict[bytes, bytes]
    client_id: str
    time: float


class StructGetBatchData(msgspec.Struct):
    keys: List[bytes]


class StructDeleteBatchData(msgspec.Struct):
    keys: List[bytes]
    client_id: str
    time: float



# Copyright (c) 2023 K.Y. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from __future__ import annotations

from pathlib import Path

from .. import FlaxKV, LevelDBDict, LMDBDict


class DBManager:
    def __init__(self, root_path="./FLAXKV_DB", raw_mode=True, log_level="DEBUG"):
        self._db_dict = {}
        self.subscribers = {}
        self._raw_mode = raw_mode
        self._root_path = Path(root_path)
        self._log_level = log_level
        if not self._root_path.exists():
            self._root_path.mkdir(parents=True, exist_ok=True)

    def detach(self, db_name: str):
        return self._db_dict.pop(db_name, None)

    def set_db(self, db_name: str, backend: str, rebuild: bool):
        self._db_dict[db_name] = FlaxKV(
            db_name=db_name,
            root_path_or_url=self._root_path.__str__(),
            backend=backend,
            rebuild=rebuild,
            raw=self._raw_mode,
            log=self._log_level,
        )

    def get(self, db_name: str, raise_key_error=False) -> LMDBDict | LevelDBDict | None:
        if raise_key_error:
            return self._db_dict[db_name]
        else:
            return self._db_dict.get(db_name, None)



# Copyright (c) 2023 K.Y. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pickle

import msgpack
import msgspec.msgpack
import numpy as np

# or use msgpack_numpy: https://github.com/lebedov/msgpack-numpy  install: pip install msgpack-numpy
# usage:
# import msgpack
# import msgpack_numpy as m
# m.patch()


class NPArray(msgspec.Struct, gc=False, array_like=True):
    dtype: str
    shape: tuple
    data: bytes


np_encoder = msgspec.msgpack.Encoder()
np_decoder = msgspec.msgpack.Decoder(type=NPArray)

try:
    import pandas as pd

    def check_pandas_type(obj):
        return isinstance(obj, pd.DataFrame)

except ImportError:

    def check_pandas_type(obj):
        return type(obj).__name__ == "DataFrame"


# def check_ext_type(obj):
#     return isinstance(obj, (tuple, set))


def encode_hook(obj):
    if isinstance(obj, np.ndarray):
        return msgspec.msgpack.Ext(
            1,
            np_encoder.encode(
                NPArray(dtype=obj.dtype.str, shape=obj.shape, data=obj.data)
            ),
        )
    else:
        return msgspec.msgpack.Ext(2, pickle.dumps(obj))
    return obj


def ext_hook(type, data: memoryview):
    if type == 1:
        serialized_array_rep = np_decoder.decode(data)
        return np.frombuffer(
            serialized_array_rep.data, dtype=serialized_array_rep.dtype
        ).reshape(serialized_array_rep.shape)
    elif type == 2:
        return pickle.loads(data.tobytes())
    return data


msg_encoder = msgspec.msgpack.Encoder(enc_hook=encode_hook)


def encode(obj, use_pickle=False):
    if use_pickle:
        return pickle.dumps(obj)
    return msg_encoder.encode(obj)


msg_decoder = msgspec.msgpack.Decoder(ext_hook=ext_hook)


def decode(data, use_pickle=False):
    if use_pickle:
        return pickle.loads(data)
    return msg_decoder.decode(data)


def decode_key(value, use_pickle=False):
    if use_pickle:
        return pickle.loads(value)
    return msgpack.unpackb(value, use_list=False)



# Copyright (c) 2023 K.Y. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from __future__ import annotations

import io
import os
import re
import shutil
import threading
import time
import traceback
from pathlib import Path
from typing import Dict, Literal
from uuid import uuid4

import msgspec
from loguru import logger

from .decorators import retry
from .pack import decode, decode_key, encode


class StructUpdateData(msgspec.Struct):
    type: str
    data: Dict[bytes, bytes]
    time: float


class DBManager:
    def __init__(
        self, db_type: str, root_path_or_url: str, db_name: str, rebuild=False, **kwargs
    ):
        """
        Initializes the database manager.

        Args:
            db_type (str): Type of the database ("lmdb", "leveldb", "remote").
            root_path_or_url (str): Root path or URL of the database.
            db_name (str): Name of the database.
            rebuild (bool, optional): Whether to create a new database. Defaults to False.
        """
        self.db_type = db_type.lower()
        self.db_name = db_name
        self._rebuild = rebuild

        url_pattern = re.compile(r'^(http://|https://|ftp://)')
        if url_pattern.match(root_path_or_url):
            self.db_address = root_path_or_url
            self.db_name = f"{db_name}-{kwargs.get('backend', 'leveldb')}"
        else:
            self.db_address = os.path.join(
                root_path_or_url, f"{db_name}-{self.db_type}"
            )

            root_path = Path(root_path_or_url)
            if not root_path.exists():
                root_path.mkdir(parents=True, exist_ok=True)

            if self._rebuild:
                self.destroy()

        self.env = self.connect(**kwargs)

    def connect(self, **kwargs):
        """
        Connects to the specified database.

        Returns:
            env: A database environment object based on the specified DB type.
        """
        if self.db_type == "lmdb":
            import lmdb

            env = lmdb.open(
                self.db_address,
                max_dbs=kwargs.get('max_dbs', 1),
                map_size=kwargs.get('map_size', 2 * 1024**3),
            )

        elif self.db_type == "leveldb":
            import plyvel

            env = plyvel.DB(self.db_address, create_if_missing=True)

        elif self.db_type == "remote":
            env = RemoteTransaction(
                base_url=self.db_address,
                db_name=self.db_name,
                backend=kwargs.pop("backend", "leveldb"),
                rebuild=self._rebuild,
                timeout=kwargs.pop(
                    "timeout", 10
                ),  # `timeout` refers to connection timeout
                **kwargs,
            )
        else:
            raise ValueError(f"Unsupported DB type {self.db_type}.")
        return env

    def rmtree(self):
        """
        Deletes the database at the specified path.
        """
        shutil.rmtree(self.db_address, ignore_errors=True)

    def destroy(self):
        """
        Destroys the database by closing and deleting it.
        """
        try:
            self.close()
        except:
            pass
        self.rmtree()
        logger.info(f"Destroyed database at {self.db_address}.")

    def rebuild_db(self):
        """
        Clears the database by closing and recreating it.
        """
        try:
            self.close()
        except:
            pass
        self.rmtree()
        self.env = self.connect()

    def get_env(self):
        """
        Retrieves the database environment.

        Returns:
            env: The database environment object.
        """
        return self.env

    def new_static_view(self):
        """
        Creates a new static view of the database.

        Returns:
            static_view: A static view of the database based on the specified DB type.
        """
        if self.db_type == "lmdb":
            return self.env.begin()
        elif self.db_type == "leveldb":
            return self.env.snapshot()
        elif self.db_type == "remote":
            return self.env
        else:
            raise ValueError(f"Unsupported DB type {self.db_type}.")

    def close_static_view(self, static_view):
        """
        Closes the provided static view of the database.

        Args:
            static_view: The static view to be closed.
        """
        if self.db_type == "lmdb":
            return static_view.abort()
        elif self.db_type == "leveldb":
            return static_view.close()
        elif self.db_type == "remote":
            return static_view.close()
        else:
            raise ValueError(f"Unsupported DB type {self.db_type}.")

    def write(self):
        """
        Initiates a write transaction on the database.

        Returns:
            wb: A write transaction object based on the specified DB type.
        """
        if self.db_type == "lmdb":
            return self.env.begin(write=True)
        elif self.db_type == "leveldb":
            return self.env.write_batch()
        elif self.db_type == "remote":
            return self.env
        else:
            traceback.print_exc()
            raise ValueError(f"Unsupported DB type {self.db_type}.")

    def close(self):
        """
        Closes the database connection.
        """
        if self.db_type == "lmdb":
            return self.env.close()
        elif self.db_type == "leveldb":
            return self.env.close()
        elif self.db_type == "remote":
            return self.env.close()
        else:
            raise ValueError(f"Unsupported DB type to {self.db_type}.")


class RemoteTransaction:
    def __init__(
        self,
        base_url: str,
        db_name: str,
        backend="leveldb",
        rebuild=False,
        timeout=10,
        http2=True,
        **kwargs,
    ):
        import httpx

        self.client = kwargs.pop(
            "client",
            httpx.Client(
                base_url=base_url,
                timeout=httpx.Timeout(None, connect=timeout),
                http2=http2,
            ),
        )
        self.db_name = db_name
        self._backend = backend
        self._rebuild = rebuild
        self.client_id = str(uuid4())

        self.put_buffer_dict = {}
        self.delete_buffer_set = set()

    @retry(
        max_retries=3,
        delay=0.2,
        backoff=2,
    )
    def attach_db(self, event: threading.Event):

        with self.client.stream(
            "POST",
            f"/connect",
            json={
                "db_name": self.db_name,
                "backend": self._backend,
                "rebuild": self._rebuild,
                "client_id": self.client_id,
            },
        ) as r:
            from httpx import Response

            r: Response
            if r.is_success:
                event.set()

            buffer = bytearray()
            chunk_size = 1024 * 1024
            try:
                for chunk in r.iter_raw(chunk_size=chunk_size):
                    if chunk == b"data: end\n\n":
                        print(f"{buffer=}")
                        yield msgspec.msgpack.decode(
                            bytes(buffer), type=StructUpdateData
                        )
                        buffer = bytearray()
                    else:
                        buffer.extend(chunk)
            except Exception as e:
                # traceback.print_exc()
                yield None

    def close_connection(self):
        url = f"/disconnect?client_id={self.client_id}"
        try:
            response = self.client.get(url)
            if not response.is_success:
                raise ValueError(response.json())
        except Exception as e:
            ...

    @retry(max_retries=3, delay=0.2, backoff=2)
    def detach_db(self, db_name=None):
        if db_name is None:
            db_name = self.db_name

        url = f"/detach"
        response = self.client.post(url, json={"db_name": db_name})
        if not response.is_success:
            raise ValueError
        return response.json()

    @retry(max_retries=3, delay=0.2, backoff=2)
    def check_db_exist(self):
        response = self.client.get(f"/check_db?db_name={self.db_name}")
        content = response.json()
        assert response.is_success, f"error: {content}"
        assert content is True

    @retry(max_retries=3, delay=0.2, backoff=2)
    def put(self, key: bytes, value: bytes):
        self.put_buffer_dict[key] = value

    def delete(self, key: bytes):
        self.delete_buffer_set.add(key)

    @retry(max_retries=3, delay=0.5, backoff=2)
    def _put_batch(self):
        byte_data = encode(
            {
                "data": self.put_buffer_dict,
                "client_id": self.client_id,
                "time": time.time(),
            }
        )
        with io.BytesIO(byte_data) as f:
            url = f"/set_batch_stream"
            files = {'file': (self.db_name, f)}
            response = self.client.post(url, files=files)

        if not response.is_success:
            raise RuntimeError
        self.put_buffer_dict = {}

    @retry(max_retries=3, delay=0.5, backoff=2)
    def _delete_batch(self):
        url = f"/delete_batch?db_name={self.db_name}"
        # todo: use stream
        response = self.client.post(
            url,
            content=encode(
                {
                    "keys": list(self.delete_buffer_set),
                    "client_id": self.client_id,
                    "time": time.time(),
                }
            ),
        )

        if not response.is_success:
            raise RuntimeError
        self.delete_buffer_set = set()

    def get(self, key: bytes, default=None):
        url = f"/get?db_name={self.db_name}"
        response = self.client.post(url, content=key)
        if not response.is_success:
            raise RuntimeError
        raw_data = response.read()
        if raw_data == b"iamnull123":
            return default
        return raw_data

    # def get_batch(self, keys: list[bytes]):
    #     url = f"/get_batch?db_name={self.db_name}"
    #     response = self.client.post(url, content=encode({"keys": keys}))
    #     if not response.is_success:
    #         raise RuntimeError
    #     raw_data = response.read()
    #     return decode(raw_data)  # non bytes

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.put_buffer_dict:
            self._put_batch()
        if self.delete_buffer_set:
            self._delete_batch()
        if exc_type is not None:
            traceback.print_exception(exc_type, exc_val, exc_tb)
            return False

    def close(self):
        """
        Do nothing.
        """

    def stat(self):
        url = f"/stat?db_name={self.db_name}"
        response = self.client.get(url)
        if not response.is_success:
            raise RuntimeError
        return decode(response.read())
